#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <iostream>
#include <vector>
#include <string>
#include "partitions.h"

using namespace std;

int main(int argc, char *argv[]){
  
  int recursive = 1;
  int pr = 0;
  string name = argv[0];  

  int bflag, ch, fd;

  bflag = 0;
  while ((ch = getopt(argc, argv, "np")) != -1) {
    switch (ch) {
    case 'n':
      recursive = 0;
      break;
    case 'p':
      pr = 1;
      break;
    case '?':
    default:
      cout << "Only options '-n' (non-recursive) and '-p' (print all partitions) are allowed" << endl;
    }
  }
  argc -= optind;
  argv += optind;

  if(argc < 2) {
    cout << name << " N k, where N is the number to partition"
      " and k is the partition size." << endl;
    exit(-1);
  }

  if(argv[0][0] == '-' || argv[1][0] == '-'){
    cout << "Please, please, positive numbers..." << endl;
    exit(-1);
  }

  u_long size = atoi(argv[1]);
  u_long n    = atoi(argv[0]);

  if(size == 0 || n == 0){
    cout << "Please specify POSITIVE! NUMBERS" << endl;
    exit(-1);
  }


  vector<u_int> partition;
  try {
    partition = vector<u_int>(size);
  } catch(bad_alloc){
    cout << "Sorry, there is not enough memory." << endl;
    exit(-1);
  }


  if(!recursive) goto non_recursive; // and proud of it :)
  

 recursive:
  //algorithm explained below at 'partitions()'
  u_int i = partitions(n, size, partition, pr);

  cerr << i << endl;
  return(0);


 non_recursive:

  boot(partition, n, size);
  u_int minimum = ceil(1.0*n/((1.0)*size));
  u_int all = 0;
  u_int sum = 0;
  u_int pos = 0;

  // Non-recursive algorithm:
  // 1) start with partition (a[0], ..., a[size-1]) = 
  //       (1, 1, 1, ..., n - size + 1) [already done above, boot(...)]
  //
  // 2) while a[size-1] >= minimum DO
  //    *) sum = a[0]
  //    *) while (k < size-1 && a[k] <= *sum/(k+1))
  //       +) sum += a[++k] 
  //    Now a[k] can be decreased (and is the 'first one [k] to be so')
  //    *) a[k]--
  //    *) substitute (a[0],..., a[k]) for the 'largest' partition with
  //       (b[0],..., b[k-1], a[k]-1). 'largest' in reverse lexicographic order
  
  while((float )partition[size-1] >= minimum){
    // this is the only needed check, because ONLY when size == 1,
    // 'boot' might set partition[0] > partition[1], because there
    // is no check inside boot in this case (this is not boot's problem)
    if(partition[0]<=partition[1]){
      if(pr) print(partition);
      all++;
    }

    pos = seek(partition, &sum);
    partition[pos]--;
    sum -= partition[pos];
    boot(partition, sum, pos);
  }

  cerr << all << endl; 

  return(0);
}


// recursive. The algorithm is quite easy:
// Start with P = (1, 1, ..., 1, n - size + 1)
//
// count = 0
// while (i = a[size - 1]) > minimum DO
//   1) a[size - 1] = i
//   2) count += partitions(n - a[size - 1], size - 1)
//
// In other words,
//
// partitions(n, k) = partitions(size - 1, size - 1) +
//                    + partitions(size, size - 1) + ...
//                    + partitions(n - ceil(n/size), size - 1)
//
// In order to be able to print *each* partition, the 'partition' vector
// is a (reference) parameter. 
// As the partition is printed in this function, we need the 'pr' parameter
// (which might well be a global variable, but these are strongly frowned upon)

u_int partitions(u_int n, u_int size, vector<u_int>& partition, int pr=0){

  u_int all = 0;

  if(size == 0 || size >  n) 
    return(0);

  boot(partition, n, size);

  if(size == 1) 
    {
      if(pr) print(partition);
      return(1);
    }
    
  for(float i = partition[size-1]; i >= (float) n /float(size); i--){
    partition[size-1] = i;
    all += partitions(n - i, size - 1, partition, pr);  
  }
  
  return(all);

}

// make the 'largest' partition available for n, size:
// ex1: n = 9, size = 4
// 6 1 1 1
// ex2: n = 9, size = 3
// 7 1 1
// ex3: n = 9, size = 2
// 8 1
// ex1: n = 9, size = 1
// 9
//
// sorry for the many examples, but as 'size' is decreased, one gets
// 'large' numbers 'to the right' (at the beginning of the sequence).
u_int boot(vector<u_int>& partition, u_int n, u_int size){

   u_int s = size;
   u_int N = n;

  while(s > 0){
    if(s == 1){
      partition[0] = N;
      break;
    }
    
    if(s < partition.size()){
      partition[s-1] = min(N-s+1, partition[s]);
    } else {
      partition[s-1] = N-s+1;
    }
    N -= partition[s-1];
    s--;
  }
  
  return(0);

  // correct but recursive: segfaults for large sizes
  //
  //
  //   if(size == 1){
  //     partition[0] = n;
  //     return(0);
  //   }
  
  //   if(size < partition.size()){
  //     partition[size-1] = min(n-size+1, partition[size]);
  //   } else {
  //     partition[size-1] = n-size+1;
  //   }
  
  //   boot(partition, n - partition[size-1], size-1);

}


// commodity function
int print(vector<u_int>& partition){
  vector<u_int>::reverse_iterator riter;
  for(riter = partition.rbegin();
      riter < partition.rend();
      riter++){
    cout << *riter << " ";
  }
  cout <<endl;
}

// find the first position which can be decreased without breaking
// the 'increasing order'
int seek(vector<u_int>& partition, u_int *sum){
  u_int i = 0;
  *sum = partition[0];

  // the sought position is either size()-1 or the first one
  // 'large enough to be decreased', which means 'larger than
  // a[0] + ... + a[k]/(k+1) '. This is easier to understand
  //
  while(i<partition.size()-1 && partition[i] <= *sum/(i+1)){
    *sum += (partition[++i]);
  };
  return(i);
}

